@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2007-2008 Butterfat, LLC (http://butterfat.net)
+Copyright (C) 2007-2010 Butterfat, LLC (http://butterfat.net)
 
 Permission is hereby granted, free of charge, to any person
 obtaining a copy of this software and associated documentation
@@ -40,10 +40,16 @@ typedef struct {
   apr_array_header_t *distrusted;
   int cookie_lifespan;
   char *server_name;
+  char *auth_program;
+  char *cookie_path;
+  bool use_auth_program;
 } modauthopenid_config;
 
 typedef const char *(*CMD_HAND_TYPE) ();
 
+// determine if a connection is using https - only took 1000 years to figure this one out
+static APR_OPTIONAL_FN_TYPE(ssl_is_https) *using_https = APR_RETRIEVE_OPTIONAL_FN(ssl_is_https);
+
 static void *create_modauthopenid_config(apr_pool_t *p, char *s) {
   modauthopenid_config *newcfg;
   newcfg = (modauthopenid_config *) apr_pcalloc(p, sizeof(modauthopenid_config));
@@ -51,11 +57,14 @@ static void *create_modauthopenid_config(apr_pool_t *p, char *s) {
   newcfg->enabled = false;
   newcfg->use_cookie = true;
   newcfg->cookie_name = "open_id_session_id";
+  newcfg->cookie_path = NULL; 
   newcfg->trusted = apr_array_make(p, 5, sizeof(char *));
   newcfg->distrusted = apr_array_make(p, 5, sizeof(char *));
   newcfg->trust_root = NULL;
   newcfg->cookie_lifespan = 0;
   newcfg->server_name = NULL;
+  newcfg->auth_program = NULL;
+  newcfg->use_auth_program = false;
   return (void *) newcfg;
 }
 
@@ -65,6 +74,12 @@ static const char *set_modauthopenid_db_location(cmd_parms *parms, void *mconfig
   return NULL;
 }
 
+static const char *set_modauthopenid_cookie_path(cmd_parms *parms, void *mconfig, const char *arg) { 
+  modauthopenid_config *s_cfg = (modauthopenid_config *) mconfig; 
+  s_cfg->cookie_path = (char *) arg; 
+  return NULL; 
+} 
+
 static const char *set_modauthopenid_trust_root(cmd_parms *parms, void *mconfig, const char *arg) {
   modauthopenid_config *s_cfg = (modauthopenid_config *) mconfig;
   s_cfg->trust_root = (char *) arg;
@@ -119,6 +134,13 @@ static const char *set_modauthopenid_server_name(cmd_parms *parms, void *mconfig
   return NULL;
 } 
  
+static const char *set_modauthopenid_auth_program(cmd_parms *parms, void *mconfig, const char *arg) {
+  modauthopenid_config *s_cfg = (modauthopenid_config *) mconfig;
+  s_cfg->auth_program = (char *) arg;
+  s_cfg->use_auth_program = true;
+  return NULL;
+} 
+
 static const command_rec mod_authopenid_cmds[] = {
   AP_INIT_TAKE1("AuthOpenIDCookieLifespan", (CMD_HAND_TYPE) set_modauthopenid_cookie_lifespan, NULL, OR_AUTHCFG,
 		"AuthOpenIDCookieLifespan <number seconds>"),
@@ -130,6 +152,8 @@ static const command_rec mod_authopenid_cmds[] = {
 		"AuthOpenIDTrustRoot <trust root to use>"),
   AP_INIT_TAKE1("AuthOpenIDCookieName", (CMD_HAND_TYPE) set_modauthopenid_cookie_name, NULL, OR_AUTHCFG,
 		"AuthOpenIDCookieName <name of cookie to use>"),
+  AP_INIT_TAKE1("AuthOpenIDCookiePath", (CMD_HAND_TYPE) set_modauthopenid_cookie_path, NULL, OR_AUTHCFG, 
+		"AuthOpenIDCookiePath <path of cookie to use>"), 
   AP_INIT_FLAG("AuthOpenIDEnabled", (CMD_HAND_TYPE) set_modauthopenid_enabled, NULL, OR_AUTHCFG,
 	       "AuthOpenIDEnabled <On | Off>"),
   AP_INIT_FLAG("AuthOpenIDUseCookie", (CMD_HAND_TYPE) set_modauthopenid_usecookie, NULL, OR_AUTHCFG,
@@ -140,19 +164,31 @@ static const command_rec mod_authopenid_cmds[] = {
 		  "AuthOpenIDDistrusted <a blacklist list of identity providers>"),
   AP_INIT_TAKE1("AuthOpenIDServerName", (CMD_HAND_TYPE) set_modauthopenid_server_name, NULL, OR_AUTHCFG,
 		"AuthOpenIDServerName <server name and port prefix>"),
+  AP_INIT_TAKE1("AuthOpenIDUserProgram", (CMD_HAND_TYPE) set_modauthopenid_auth_program, NULL, OR_AUTHCFG,
+		"AuthOpenIDUserProgram <full path to authentication program>"),
   {NULL}
 };
 
-/* Get the full URI of the request_rec's request location */
-static void full_uri(request_rec *r, std::string& result, modauthopenid_config *s_cfg) {
+// Get the full URI of the request_rec's request location 
+// clean_params specifies whether or not all openid.* and modauthopenid.* params should be cleared
+static void full_uri(request_rec *r, std::string& result, modauthopenid_config *s_cfg, bool clean_params=false) {
   std::string hostname(r->hostname);
-  std::string protocol(r->protocol);
   std::string uri(r->uri);
   apr_port_t i_port = ap_get_server_port(r);
-  std::string prefix = (i_port == 443) ? "https://" : "http://";
+  std::string prefix = (using_https != NULL && using_https(r->connection)) ? "https://" : "http://";
   char *port = apr_psprintf(r->pool, "%lu", (unsigned long) i_port);
   std::string s_port = (i_port == 80 || i_port == 443) ? "" : ":" + std::string(port);
-  std::string args = (r->args == NULL) ? "" : "?" + std::string(r->args);
+
+  std::string args;
+  if(clean_params) {
+    opkele::params_t params;
+    if(r->args != NULL) params = modauthopenid::parse_query_string(std::string(r->args));
+    modauthopenid::remove_openid_vars(params);
+    args = params.append_query("", "");
+  } else {
+    args = (r->args == NULL) ? "" : "?" + std::string(r->args);
+  }
+
   if(s_cfg->server_name == NULL)
     result = prefix + hostname + s_port + uri + args;
   else
@@ -168,6 +204,11 @@ static int show_input(request_rec *r, modauthopenid_config *s_cfg, modauthopenid
   if(r->args != NULL) 
     params = modauthopenid::parse_query_string(std::string(r->args));
   modauthopenid::remove_openid_vars(params);  
+
+  std::string uri_location;
+  full_uri(r, uri_location, s_cfg, true);
+  params["modauthopenid.referrer"] = uri_location;
+
   params["modauthopenid.error"] = modauthopenid::error_to_string(e, true);
   return modauthopenid::http_redirect(r, params.append_query(s_cfg->login_page, ""));
 }
@@ -180,7 +221,7 @@ static int show_input(request_rec *r, modauthopenid_config *s_cfg) {
     params = modauthopenid::parse_query_string(std::string(r->args));
   modauthopenid::remove_openid_vars(params);
   std::string uri_location;
-  full_uri(r, uri_location, s_cfg);
+  full_uri(r, uri_location, s_cfg, true);
   params["modauthopenid.referrer"] = uri_location;
   return modauthopenid::http_redirect(r, params.append_query(s_cfg->login_page, ""));
 }
@@ -249,8 +290,11 @@ static int start_authentication_session(request_rec *r, modauthopenid_config *s_
 					std::string& return_to, std::string& trust_root) {
   // remove all openid GET query params (openid.*) - we don't want that maintained through
   // the redirection process.  We do, however, want to keep all aother GET params.
-  // also, add a nonce for security
+  // also, add a nonce for security 
   std::string identity = params.get_param("openid_identifier");
+  // pull out the extension parameters before we get rid of openid.*
+  opkele::params_t ext_params;
+  modauthopenid::get_extension_params(ext_params, params);
   modauthopenid::remove_openid_vars(params);
 
   // add a nonce and reset what return_to is
@@ -258,16 +302,15 @@ static int start_authentication_session(request_rec *r, modauthopenid_config *s_
   modauthopenid::make_rstring(10, nonce);
   modauthopenid::MoidConsumer consumer(std::string(s_cfg->db_location), nonce, return_to);    
   params["modauthopenid.nonce"] = nonce;
-  //remove first char - ? to fit r->args standard, then get the return_to url with original non-openid get parameters
-  std::string args = params.append_query("", "").substr(1); 
-  apr_cpystrn(r->args, args.c_str(), 1024);
   full_uri(r, return_to, s_cfg);
+  return_to = params.append_query(return_to, "");
 
   // get identity provider and redirect
   try {
     consumer.initiate(identity);
     opkele::openid_message_t cm; 
     re_direct = consumer.checkid_(cm, opkele::mode_checkid_setup, return_to, trust_root).append_query(consumer.get_endpoint().uri);
+    re_direct = ext_params.append_query(re_direct, "");
   } catch (opkele::failed_xri_resolution &e) {
     consumer.close();
     return show_input(r, s_cfg, modauthopenid::invalid_id);
@@ -292,11 +335,14 @@ static int start_authentication_session(request_rec *r, modauthopenid_config *s_
 static int set_session_cookie(request_rec *r, modauthopenid_config *s_cfg, opkele::params_t& params, std::string identity) {
   // now set auth cookie, if we're doing session based auth
   std::string session_id, hostname, path, cookie_value, redirect_location, args;
+  if(s_cfg->cookie_path != NULL) 
+    path = std::string(s_cfg->cookie_path); 
+  else 
+    modauthopenid::base_dir(std::string(r->uri), path); 
   modauthopenid::make_rstring(32, session_id);
-  modauthopenid::base_dir(std::string(r->uri), path);
   modauthopenid::make_cookie_value(cookie_value, std::string(s_cfg->cookie_name), session_id, path, s_cfg->cookie_lifespan); 
   modauthopenid::debug("setting cookie: " + cookie_value);
-  apr_table_setn(r->err_headers_out, "Set-Cookie", cookie_value.c_str());
+  apr_table_set(r->err_headers_out, "Set-Cookie", cookie_value.c_str());
   hostname = std::string(r->hostname);
 
   // save session values
@@ -304,7 +350,10 @@ static int set_session_cookie(request_rec *r, modauthopenid_config *s_cfg, opkel
   sm.store_session(session_id, hostname, path, identity, s_cfg->cookie_lifespan);
   sm.close();
 
+  opkele::params_t ext_params;
+  modauthopenid::get_extension_params(ext_params, params);
   modauthopenid::remove_openid_vars(params);
+  modauthopenid::merge_params(ext_params, params);
   args = params.append_query("", "").substr(1);
   if(args.length() == 0)
     r->args = NULL;
@@ -328,9 +377,16 @@ static int validate_authentication_session(request_rec *r, modauthopenid_config
       consumer.close();
       return show_input(r, s_cfg, modauthopenid::invalid_nonce); 
     }
+
+    // if we should be using a user specified auth program, run it to see if user is authorized
+    if(s_cfg->use_auth_program && !modauthopenid::exec_auth(std::string(s_cfg->auth_program), consumer.get_claimed_id())) {
+      consumer.close();
+      return show_input(r, s_cfg, modauthopenid::unauthorized);       
+    }
+
     // Make sure that identity is set to the original one given by the user (in case of delegation
     // this will be different than openid_identifier GET param
-    std::string identity = consumer.get_normalized_id();
+    std::string identity = consumer.get_claimed_id();
     consumer.kill_session();
     consumer.close();
 
@@ -362,9 +418,9 @@ static int mod_authopenid_method_handler(request_rec *r) {
   if(has_valid_session(r, s_cfg))
     return DECLINED;
 
-  // parse the get params
+  // parse the get/post params
   opkele::params_t params;
-  if(r->args != NULL) params = modauthopenid::parse_query_string(std::string(r->args));
+  modauthopenid::get_request_params(r, params);
 
   // get our current url and trust root
   std::string return_to, trust_root;
